/*
  This program is free software published under the
  terms of the GNU Lesser General Public License.
  http://www.gnu.org/copyleft/lesser.html    
  This program is distributed in the hope that it will
  be useful - WITHOUT ANY WARRANTY; without even the
  implied warranty of MERCHANTABILITY or FITNESS FOR A
  PARTICULAR PURPOSE.
*/ 


/*  xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
     source code: ES6 (ECMAScript 6, ECMAScript 2015, JavaScript 6)
    xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx */
 
//  xxxxxxxxxxxxx          version: 1.0              xxxxxxxxxxxxxxx
		 
//  xxxxxxxxxxxxxxx   global variable declarations   xxxxxxxxxxxxxxxx

 
// These variables below can be directly modified by the user on the options form: 

var item_cnt= 5;       // The number of stimuli in the task

var rep_cnt = 3;       // maximum number of tone repetitions in each item in task

var before = 10;       // delay time before call playSound() function x (clock = 100ms) => ms

var after = 10;				 // delay time before item panel is displayed x (clock = 100ms) => ms

var is_range = false;   // use all tones or only a limited range. see stimGen() function.
						//var min_range;  // Upper limit - the maximum duration of all tones in ms (limits are included)
						//var max_range;  // Lower limit - the minimum duration of all tones in ms (limits are included) 
var is_back = true;    // if (is_back==true) => user can go back to previous task/item.  

var is_mis = true;    // if (is_mis==true) => missing values are allowed

var prg_type = true;   //  type of information in Progress Bar (true => Items, false => Nothing) 

var demo = false;     // if (demo == true) => Show task solutions 
    		
var store = false;      // false => this web application will not store results within client web browser.; true => This web application will store the results inside client web browser. 

 
// Other global constants and variables:

const clock = 100;  // basic timer resolution in ms. Default = 100ms

var timer_id = null;   // timer_id is the identifier of the repeated action you want to cancel. This ID is returned from setInterval().

var is_storage= false; // local storage is enabled on user browser or not

var step = -1;   		 	 // global timer counter => see appTimer()

var first = false;    // true => play tone (sound); false => show item panel

var delay;     // timer will wait for this period. delay == before (if first == true) or after (if first == false)
   
var dirty = false;  // if  (dirty==true) => user changed any options in form 

var startDate;  // Task start time

var item = 0;   // current item in task     

// global (window) arrays
	
var stim = [];   // stimulus content (length/name of .wav files in audio dir) for each item. range dimensions: [0..item_cnt-1] x [0..1]

var ans = [];   //  the user's response to each item in the task [0..item_cnt-1]

var repeated = [];  // for each item: the number of completed repetitions of the sound stimulus for each item. [0..item_cnt-1]


// application init [create/start Timer, check storage, show start panel..] 
function init() 
{  
  if (timer_id)  {   
    window.clearInterval(timer_id);  
  }  
  timer_id= setInterval(function () {appTimer()},clock);  
  is_storage= isStorage() && storageAvailable('localStorage');
  if (is_storage)  {
    document.getElementById('store_group').style.display= 'block';
		document.getElementById('no_storage').style.display= 'none';
  }			
	let e= document.getElementById('answer');  // attach event handlers..
	e.onkeypress = answerKeyPress;	
	e.onkeyup = answerKeyUp;	
  e.onpaste = event => false; // work in ECMAScript 2015	
		//e.onpaste= function(){return false}; // Prevent pasting in answer input type=text (since pasted content might include non-allowed characters)
	
	setIns();		
	
	showStart(); 
}


// application timer handler function
// basic timer resolution is aprox 100 ms
function appTimer()    
{    
	 if (step >= 0)  //   if (step < 0 => timer nop, idle, just wait
	 {          		
		 step++;  
		 if (step == delay) 
		 {   
			 step= -1;       // timer must wait. audio play in progress or item_div panel is visible.  nop, idle, do none, Timer no operation  			   
			 if (first)  {
				 playSound();    // play sound (wav file from audio dir)	 
       }
			 else  {  // show item panel
				 document.getElementById('item_div').style.display= 'block';
				 showItem();   // update (show) Item panel
			 }				 
		 }
	 }
} 	


// before using web storage, check browser support for localStorage and sessionStorage
function isStorage()  
{
  if (typeof(Storage) !== 'undefined') {
     return true;                 
  } 
  else {
    return false;    
  }  
}


// necessarily because of IE11 offline mode
function storageAvailable(type) 
{
  try  {
    return (window[type]) ? true: false;   
  }
  catch(e) {
    return false;
  }
}


// show start panel 
function showStart()  
{      
  document.getElementById('end').style.display= 'none';
  document.getElementById('result_wrapper').style.display= 'none';  
  document.getElementById('store_id').style.display= 'none';
	document.getElementById('table_grp').style.display= 'none';
	document.getElementById('start').style.display= 'block';	
}


// set instructions text depending on user selected task options
function setIns()  
{
  document.getElementById('instructions').style.display= 'block';		   
  if (rep_cnt > 1) {		
		document.getElementById('ins_rep_cnt').style.display = 'block';
		document.getElementById('ins_rep_cnt').textContent= 'If you did not hear the tone well in the current task, click the "Repeat" button and the tone will play again. You can use this option ' + rep_cnt + ' times within the same task.';
	}
	else  {
		document.getElementById('ins_rep_cnt').style.display = 'none';
	}
  document.getElementById('ins_mis').style.display = is_mis ? 'none' : 'block';  
	document.getElementById('ins_back').style.display = is_back ? 'block' : 'none'; 	
	document.getElementById('ins_demo').style.display = demo ? 'block' : 'none';
}	


// refresh Options (update *_msg label) => give user short option HELP
function refOpt(e)  
{   
  dirty=true;       
  let my_id = e.id, msg = document.getElementById(my_id + '_msg'), s = 'Values within this range are allowed: ', s2= ' with step= ';               
	switch (my_id)
	{
		
		case 'item_cnt':
		case 'rep_cnt':
		case 'before':
		case 'after':
			if (e.validity.valid)  {
				msg.textContent= ''; 
				msg.removeAttribute('class');
			}
			else  {
				msg.textContent= s + e.min + ".." + e.max + s2 + e.step;
				msg.setAttribute('class', 'error');
			} 
			break;  		
			
    case 'is_range':
      if (e.checked)  {
				msg.txtContent= 'Only tones of the selected length range will be used. Use if users fail to hear very short tones or if you have problems with them.';
        document.getElementById('range_div').style.display = 'block'; 
			}
			else  {
				msg.txtContent= 'All available tones (35 different tone lengths) will be included.';
        document.getElementById('range_div').style.display = 'none'; 				
			}
		  break;
			
    case 'is_back':     		
			msg.textContent= e.checked ? 'Yes' : 'No. Returning to previous items is not allowed.';    
			break;
    
    case 'is_mis':
    	msg.textContent= e.checked ? 'Missing values are allowed. The user does not have to enter a tone length value.': 'Missing values are not allowed. The user must enter a valid tone length value before moving to the next item. Range: [1..100000ms = 0.001..100s].';    
			break;
			    
    case 'prg_type':                                                                                                                                                  
      msg.textContent= e.checked ? 'Progress bar will show the number of the current item and the number of remaining items until the end of the task.' : 'Progress bar will be invisible';     
      break;    
				
		case 'demo':
			msg.textContent= e.checked ? 'The correct solution will be displayed on each item. At the end of the task, the first table will show the exact solutions for each item.' : 'No';  
			break; 
		
		case 'store':                                                                                                                                            
			msg.textContent= (e.value === '1') ? 'This web application will store the results inside your web browser (no expiration date). Useful for group research' : 'This web application will not store results within your web browser. Data is lost when the browser tab is closed';
			break;
 
	}   // switch
	
  refMsg();   
}


// refresh main validator message for user defined OPTIONS 
function refMsg()   
{ 	
	if ( optionsValid() )  {
    document.getElementById('is_valid').textContent=  dirty == true ? 'All your settings have been successfully updated' : '';  
    document.getElementById('is_valid').setAttribute('style', 'text-align:center;color:#000000'); 
  }
  else  {
    document.getElementById('is_valid').textContent='Some of your options are incorrectly set. They were not updated.';   // These settings are marked in red. 
    document.getElementById('is_valid').setAttribute('style', 'text-align:center;color:#ff0000;font-weight:bold'); 
  } 	
}


// check for validity.valid object in all input type="number"
function optionsValid()  
{     
  let res= document.getElementById('item_cnt').validity.valid && document.getElementById('rep_cnt').validity.valid
				&& document.getElementById('before').validity.valid && document.getElementById('after').validity.valid;	
  return res;              
}


// start task (research) 
function startTask(e)   
{     
  initLayout();  
	
	setPar();
 
	stimGen();
		   
	showItem();     // show div/panel for first item
   		
}


// init page layout when task is started. init visible/invisible html dom elements, progress bar
function initLayout()
{
	document.getElementById('result_wrapper').style.display= 'none';
  document.getElementById('end').style.display= 'none'; 
  document.getElementById('start').style.display= 'none';              
  document.getElementById('task_section').style.display= 'block'; 
	
	document.getElementById('nav_items').textContent= '';	
  if (!prg_type)  {
		document.getElementById('progress_holder').style.display= 'none';    	
	}	
	else  {
		document.getElementById('progress_holder').style.display= 'block';   				   			
		document.getElementById('progress').setAttribute('style', 'width:100%');  // reset progress bar width						 	
	} 
	document.getElementById('demo_sol').style.display= demo ? 'block' : 'none';
}


// set internal application research parameters (global vars)
function setPar()  
{  
  stim= []; 
  ans= []; 	
	repeated= [];
	for (let i= 0; i< item_cnt; i++)   {  // i => item        
    ans[i]= '';	
		repeated[i]= 0;
  } 
  item= 0;    	
  startDate = new Date();			
	first= true; 
	delay= before; 
}


// generates task stimuli. stimulus generator
function stimGen()	
{
	const files= [ 0.01, 0.02, 0.03, 0.04, 0.05, 0.06, 0.07, 0.08, 0.09, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1, 1.1, 1.2, 1.3, 1.4, 1.5, 1.6, 1.7, 1.8, 1.9, 2, 2.5, 3, 3.5, 4, 4.5, 5 ];
  let len= files.length, lb, hb, f, s, ok;  // default files.length = 35
  
	if (is_range) {
		lb= parseInt(document.getElementById('min_range').selectedIndex, 10);  // parseInt is probably redundant?
	  hb= len - 1 - parseInt(document.getElementById('max_range').selectedIndex, 10);
	}
	else  {
		lb= 0;
		hb= len-1;
	}
	
	// full range is: files[Math.floor(Math.random() * len)];   // index returns a random integer from 0 to files.length-1
	for (let i= 0; i< item_cnt; i++)   
	{  // i => item  
		do 
		{			
			f= files[getRndInt(lb,hb)];  // name of audio file. see function playSound()
			if (i == 0) { 
				ok= true; 
			}
			else	{  // only this condition: selected file in i-iteration must be different from file in i-1 iteration
				ok= ( f!= stim[i-1] );  
			}	        
		} while (!ok); 
		stim[i]= f;  // selected tone (audio wav file)
  }	
}


// get random integer from min to max range
function getRndInt(min, max) {
	return Math.floor(Math.random() * (max - min + 1)) + min;
}


// Changes application settings that can be changed directly by the user within the options form
function setVarPar()  
{   
	
	if (document.getElementById('item_cnt').validity.valid)  {  
	  item_cnt= parseInt(document.getElementById('item_cnt').value, 10);			
  } 
	 
  if (document.getElementById('rep_cnt').validity.valid) {
		rep_cnt= parseInt(document.getElementById('rep_cnt').value, 10);	
  }
	 
  if (document.getElementById('before').validity.valid) {
		before= Math.round(parseInt(document.getElementById('before').value, 10)/clock);
  }
  
	if (document.getElementById('after').validity.valid) {
		after= Math.round(parseInt(document.getElementById('after').value, 10)/clock);  
  }

	is_range= document.getElementById('is_range').checked ? true: false; 

	is_back= document.getElementById('is_back').checked ? true: false;       
  
  is_mis= document.getElementById('is_mis').checked ? true: false; 
	
	prg_type= document.getElementById('prg_type').checked ? true : false;  // display item count on footer or nothing
		
	demo= document.getElementById('demo').checked ? true: false;	
	
  store= (document.getElementById('store_group').disabled === false && document.getElementById('store').value === '1') ? true : false;
  
}


// goto previous item (stimulus)
function priorItem()  
{  
  ans[item]= trim(document.getElementById('answer').value);
	//ans[item]= parseInt(trim(document.getElementById('answer').value),10);
	if (item > 0)  {
    item--;  
    showItem();     // draw again previous item 
  }         
}


// goto next item (stimulus)
function nextItem()  
{ 
  ans[item]= trim(document.getElementById('answer').value);
	
	if (!valAnswer(document.getElementById('answer').value ) )  {  
		alert('You must enter a valid tone duration value before moving on to the next task. \n  You must enter valid integer value between 1 and 100000 ms'); 
		return;
  }	

	item++;					
	if (item < item_cnt)  {         		 
    showItem();     // show next item => range 0..item_cnt-1        
  }
  else  {               
    theEnd();  // end of task if item >= item_cnt. Note: only this code line calling theEnd() function 
  }
}


// trim input string (user answer)
function trim(s)  {
	return s.replace(' ', '');
}
 
 
// validate user answer. par: v => user answer value for current item
function valAnswer(v)  
{	 
	v= trim(v);  
	if ( is_mis && v == '') {
    return true; // missing values are allowed
  }
  else 
	{	
		let ok= true;
		if ( v == '' || isInt(v) == false || parseInt(v,10) < 1 || parseInt(v,10) > 100000 )   {  
			ok= false;  // missing value is also invalid
		}	
		document.getElementById('ans_val').textContent= ok ? '' : ' You must enter valid numeric value between 1 and 100000 ms';
		return ok;
	}  
}


// Checks whether it is an integer value or not
// isNaN("") = false: the empty string is converted to 0 which is not NaN                   
function isInt(n) {       
   return (!isNaN(n)) && (parseFloat(n) === parseInt(n, 10)) &&  (n % 1 === 0); 
} 


/* see function init().  deprecated but still works:
   https://developer.mozilla.org/en-US/docs/Web/API/GlobalEventHandlers/onkeypress
   this regex [1-9()*\/+-] was tested here:    https://regex101.com/    */
function answerKeyPress(event)   { 
  return event.charCode === 0 || /[0-9]/.test(String.fromCharCode(event.charCode));	
}


/* go to the next task if the ENTER or PageDown key is pressed. 
   go to the previous task if the PageUp key is pressed.
   see function init()   */
function answerKeyUp(e)  
{  
	if ( e.key === 'Enter' || e.keyCode === 13 ) {  
		nextItem();
	}
}


// raise when entered answer changes. see html tag
// par => value => user answer value ( document.getElementById('answer').value )
function answerInput(value)  
{	
  let a = trim(value);   // user answer (trimed)
	let s = document.getElementById('ans_sec');   // label for convert ms into sec (only if valid entry)
	if ( a == '' )  { 
	  s.textContent= ''; 		// clear if no user entry 
	} 	
	else if ( isInt(a) )  {
		let n= parseInt(a,10) / 1000;
		if ( n >= 0.001 && n <= 100 )  {
			s.textContent=  ' = ' + n.toFixed(3) + ' seconds';
		}
		else  {
			s.textContent= ' Invalid value'; 
		}
	}
	else {
		s.textContent= ' Invalid value';	
	}
	//s.textContent= 'You must enter a valid tone duration value before moving on to the next task.';  		
  //s.textContent= ' Invalid value';
}


// show current item_div panel/div. items are in range: 0..item_cnt-1     
function showItem() 
{    		
  if (repeated[item] == 0)  {  // only first time when item is displayed	
	  repeatSound();  // hide item_div panel and start timer with first= true
		return;
	}
	
	// update footer (navigation buttons, progress bar,..)
	document.getElementById('prior').style.display= (item > 0 && is_back) ? 'inline': 'none';  // set footer navigation prior button (left arrow) 		
	 
	let btn= document.getElementById('next');     
	if(item == item_cnt-1)  {       
		btn.style.backgroundImage=  "url('img/end.png')";  // update "Next item" image on "right arrow" button 
		btn.title= 'End of task';
	}
	else  {
		btn.style.backgroundImage= "url('img/forward.png')"; 
		btn.title= 'Next';
	} 
					
	if (prg_type)  {    // update Progress bar => show current item count and remaining item count on Progress Bar (default)                                                                                                                                                       
		let percent = Math.round(100*(item_cnt-(item+1))/item_cnt);      
		document.getElementById('progress').setAttribute('style', 'width:'+percent+'%');    // !important
		document.getElementById('nav_items').textContent= 'Item= ' + (item + 1) + ' Remaining= ' + (item_cnt - item-1);  
	}
	
  document.getElementById('repeat').style.display= repeated[item] < rep_cnt ? 'block' : 'none';  // enable or disable repeating tone
	
	// refresh string values for current item
	document.getElementById('answer').value= ans[item];	   // user answer
  answerInput(document.getElementById('answer').value);  // refresh convert into seconds
		//document.getElementById('ans_val').textContent= '';    // tmp clear validator (until Next btn is clicked) 
	
	valAnswer(document.getElementById('answer').value);  // refresh user answer validator
		
	document.getElementById('answer').focus();		
	
  document.getElementById('demo_sol').textContent= 'Correct solution = ' + (1000 * stim[item]);  
}


// play only one sound stimulus. play audio from ../audio/fix dir. 
function playSound()
{   
	let sound= document.getElementById('sound');		
  let file = 'audio/' + stim[item] + '.wav';
	sound.setAttribute('src', file);  // sound.load();
	sound.play();
	sound.onended = function() {
		audioEnded();    
	}; 			  
} 


// audio ended => when the audio file is finished
function audioEnded()
{  
  first= false;
	delay= after;   // update timer delay period
	repeated[item]++;
	step= 0;  // reset timer => enable timer for new delay period (after) 
} 


// play again first and second tone in current item.
// this event is trigger by user click
function repeatSound()
{
	document.getElementById('item_div').style.display= 'none';
	if ( repeated[item] > 0 ) {
		ans[item]= document.getElementById('answer').value; 
	}
	first= true; 
	delay= before;	
	step= 0;   // activate timer again. timer is on
}


// end of task (research)
function theEnd()  
{   
	step= -1;  // timer nop. probably redundant
	
	document.getElementById('task_section').style.display= 'none';	
  document.getElementById('end').style.display= 'block';   
  document.getElementById('result_wrapper').style.display= 'block';
		
	let mis= 0, is_mis, cor, err, errors= [], ans_v= [], stim_v= [], rows = [], tmp;
  let header = demo ? ['Item', 'Estimation', 'Length', 'Error %'] : ['Item', 'Estimation', 'Error %']	
	const D= 5;  // number of decimals 
	for (let i= 0; i< item_cnt; i++)   
	{
		if ( ans[i] == '' || !isInt(ans[i]) || !isFinite(ans[i]) || parseInt(ans[i],10) < 1 || parseInt(ans[i],10) > 100000) {
      is_mis= true;
		}
    else {
      is_mis= false;
		}			
		cor = 1000 * stim[i]; // correct answer in ms. user answer is in ms. So we have to convert stim from seconds into ms
		if ( is_mis )  {  
		  mis++;		// mis = number of missing items in task 
    }		
		else  {  	  
			tmp= parseInt(ans[i],10);  // convert string to integer
			err= 100 * Math.abs(tmp - cor) / cor;      // errors for this i-item (%)
			errors.push(err); // errors => array of all errors in task (missing values are excluded)   
			stim_v.push(cor);  // valid (non-missing) values od correct tone length in ms		
			ans_v.push(tmp);	 // valid (non-missing) user estimation of tone length in ms
		}					
		if (demo) {
			rows[i]= is_mis ? [ parseInt(i+1,10), '', cor, '' ] : [ parseInt(i+1,10), ans[i], cor, err.toFixed(D) ]; 
		}
		else  {
		  rows[i]= is_mis ? [ parseInt(i+1,10), '', '' ] : [ parseInt(i+1,10), ans[i], err.toFixed(D) ]; 
		}
	}			
	createTableArray(document.getElementById('table1'), rows, header);	
	
	// calculate decriptive statistics for the errors (%) => table2  	  
	rows = [];
	let n= errors.length;
	let m = ( n > 0 ) ? Stat.m(errors).toFixed(D) : '';      // mean only for "valid" items in task
	let sd = ( n > 1 ) ? Stat.sd(errors, m).toFixed(D): '';  // standard deviation only for valid items in task  	
		 
	header=  ['Variable', 'Average', 'Sample Standard Deviation', 'N']; // header
	rows.push( ['Error (%)', m, sd, n] );  // last row
	
	createTableArray(document.getElementById('table2'), rows, header);  
	
	// calculate inferential statistics (pearson r, sperman ro, t-arguments, t-distribution probabilities)	
	header=  ['Variables', 'Spearman-ro', 't-value ro', 'df', 'p-ro (two-sided)',  'Pearson r', 't-value r', 'df', 'p-r (two-sided)' ]; // header 
	rows=[];
	for (let j= 0; j<2; j++)
	{
		let y=[], vars;
		if ( j == 0 )  {
			y= ans_v;
			vars= 'Length-Estimation';
		}
		else  {
			y= errors;
			vars= 'Length-Errors';
		}	
		 
		let row= [ vars, '', '', '', '', '', '', '', '' ];  // range 0..8
		if  ( n >= 3 ) 
		{			
			let ro= Stat.r_spearman(stim_v, y);     // spearman's rank correlation
			if (!isNaN(ro) && isFinite(ro) )  // if no variance in both arrays (sd= 0) => Infinity or Nan
			{ 
				row[1]= ro.toFixed(D);  
				let t_ro= ro * Math.sqrt ( ( n-2 ) / (1 - ro * ro) );  // t-argument for spearman correlation 		
        if (!isNaN(t_ro) && isFinite(t_ro))  
				{
					let err=[0];  // reset error flag indicator (must be 0)
					let p_ro= Stat.prncst ( Math.abs(t_ro), n-2, 0, err[0]); // t-argument, df, 0, error
					p_ro=  2 * (1 - p_ro);  // Propability p (two-sided) => tested ok for agument n= 3..1000 with: https://www.psychometrica.de/correlation.html 
					row[2]= t_ro.toFixed(D);
					row[3]= n-2;
					if (err[0] == 0) { 
						row[4]= p_ro.toFixed(D); // propability Spearman ro (two-sided)
					}
				}
			}
			
			let r= Stat.r(stim_v, y);   // Pearson r correlation
			if (!isNaN(r) && isFinite(r))    // if no variance in both arrays (sd= 0) => Infinity or Nan
			{
				row[5]= r.toFixed(D);  
				let t_r=  r * Math.sqrt ( ( n-2 ) / (1 - r * r) );  // t-argument for pearson correlation
				if (!isNaN(t_r) && isFinite(t_r)) 
				{					
					err=[0];  // error flag indicator (must be 0)
					let p_r= Stat.prncst ( Math.abs(t_r), n-2, 0, err[0]);  // t-argument, df, 0, error
					p_r= 2 * (1 - p_r);  // Propability p (two-sided) => tested ok for agument n= 3..1000 with: https://www.psychometrica.de/correlation.html 
					row[6]= t_r.toFixed(D);
					row[7]= n-2;
					if (err[0] == 0) { 
						row[8]= p_r.toFixed(D);  // p-r (two-sided)
					}					 
				}
			}			
		}  //  if ( n >= 3 ) 
		
		rows.push( row );  // 9 x columns
	}		
	createTableArray(document.getElementById('table3'), rows, header); 
  document.getElementById('inf_warning').style.display= n >= 3 ? 'block' : 'none';	
	
	// update 'score' label => above "Repeat" button
	switch (n)  {
		case 0: 
			document.getElementById('score').textContent= 'Descriptive statistics cannot be calculated. Missing items= ' + mis;
			break;
		case 1: 
			document.getElementById('score').textContent= 'Average error= ' + m + '%. Standard deviation cannot be calculated. Missing items= ' + mis;
			break;	
		default:
		document.getElementById('score').textContent= 'Average error= ' + m + '%. Sample Standard Deviation= ' + sd + '%. Missing items= ' + mis; 
	}
	
	document.getElementById('l_ts').textContent='Last completed research: ' + new Date();
  document.getElementById('l_res').style.display= 'block';
		
	try 
  {
    if (store) {
			setStorage(m, sd, mis, rows[0]);  
    }
  }        
  catch (e)  { 
    document.getElementById('table_grp').style.display= 'block';
		document.getElementById('table_grp').textContent= "Your browser probably doesn't have support for the Web Storage Object or access to the Storage Object is disabled. Check your browser settings. In some browsers (IE11, Edge) Storage Object is not enabled when accessed in a local file (offline)."; 
  }   	 
 	
}


// return integer into string hh:mm:ss time format
// input par: total == integer (time in seconds)
function stringTime(total) 
{
  let hh = Math.floor(total/(60*60));         
  let div_min = total % (60 * 60);
  let mm = Math.floor(div_min / 60);       
  let div_sec = div_min % 60;
  let ss = Math.ceil(div_sec);    
  // These lines ensure you always have two-digits. This formats your string to HH:MM:SS
  if (hh < 10)  {hh = "0"+hh;}     
  if (mm < 10) {mm = "0"+mm;}
  if (ss < 10) {ss = "0"+ss;} 
  return hh + ":" + mm + ":" + ss;
} 

 
/* Store the results of the last completed task in the user's browser
   set user data into window.localStorage. Parameters:	 
	 m = variable errors. mean only for valid items in task
   sd = variable errors. standard deviation only for valid items in task
	 mis = number of missing items in task  
	 data => inferential statistics for Length-Estimation variables (2. row of table3) */
function setStorage(m, sd, mis, data)   
{   
	let endDate= new Date();
	let tt= Math.round( (endDate.getTime() - startDate.getTime() ) / 1000);  // task time 
	tt= stringTime(tt); 
	
	let oldItems = JSON.parse(localStorage.getItem('tde-v1.0')) || []; 
  let id= oldItems.length+1;     	
	// set new "record"
  let newItem = {
      'id' : id,
      'date' : setDate(),   //new Date(),  		   
      'mean' : m,
			'sd': sd,
			'missing' : mis,
			'ro': data[1],  // Spearman-ro
			't-ro': data[2], // t-value ro
			'p-ro': data[4],  // p-ro (two-sided)	
			'r': data[5],  // Pearson r
			't-r': data[6], // t-value r	
			'p-r': data[8],  // p-r (two-sided)
			'df': data[7],   // df
			
			'demo': demo ? 1 : 0,
			'items': item_cnt,						
			'repeated': rep_cnt,			
	    'time-before' : before * clock,        
      'time-after' : after * clock,			
			'low-limit': is_range ? document.getElementById('min_range').value : '',
			'up-limit' : is_range ? document.getElementById('max_range').value : '',
			'backward': is_back ? 1 : 0,
			'required' : is_mis ? 0 : 1,  			
      'task-time': tt			
  };

  oldItems.push(newItem);

  localStorage.setItem('tde-v1.0', JSON.stringify(oldItems));
  
	document.getElementById('table_grp').style.display= 'block';
  document.getElementById('store_id').style.display= 'block';
  document.getElementById('store_id').textContent= 'The results are stored in your browser. ID= ' + id ;
  
	createTableJSON(document.getElementById('table_grp'), oldItems);  
 
}


// Formats the date in the desired format. If you do not want this string format do not call this function. use: new Date()
function setDate()  
{
	let today = new Date();
	let dd = today.getDate();
	let mm = today.getMonth()+1; 
	let yyyy = today.getFullYear();
	//if(dd<10) { dd='0'+dd; } 
	if(mm<10) {	mm='0'+mm; } 
	today = dd+'.' + mm + '.' +yyyy;
	return today;
}


// convert JSON data dynamically to HTML Table using JavaScript - Dynamically Create table with rows in JavaScript
// downloaded and adapted from source: https://www.encodedna.com/javascript/populate-json-data-to-html-table-using-javascript.htm
function createTableJSON(elem, my_json) 
{  
  // EXTRACT VALUE FOR HTML HEADER. 
  let col = [];
  for (let i = 0; i < my_json.length; i++) {
    for (let key in my_json[i]) {
        if (col.indexOf(key) === -1) {
            col.push(key);
        }
    }
  } 
  let table = document.createElement("table");
  // CREATE HTML TABLE HEADER ROW USING THE EXTRACTED HEADERS ABOVE.
  let tr = table.insertRow(-1);                   // TABLE ROW.
  for (let i = 0; i < col.length; i++) {
      let th = document.createElement("th");      // TABLE HEADER.
      th.innerHTML = col[i];
      tr.appendChild(th);
  }
  // ADD JSON DATA TO THE TABLE AS ROWS.
  for (let i = 0; i < my_json.length; i++) 
  {
    tr = table.insertRow(-1);
    for (let j = 0; j < col.length; j++) {
      let tabCell = tr.insertCell(-1);
      tabCell.innerHTML = my_json[i][col[j]];
    }
  }
  // FINALLY ADD THE NEWLY CREATED TABLE WITH JSON DATA TO A CONTAINER.
  elem.innerHTML= "";
  elem.appendChild(table);
}


/* convert array data dynamically to HTML table - dynamically create html table
   elem == DOM html element (node)
   data => 1-dim or 2-dim array (matrix)
   col => string column descriptions. Set an empty string if you want to skip a column inside a 2-dim array	 */ 
function createTableArray(elem, data, col) 
{  
	let table= document.createElement('table');  // html table
  let tr= table.insertRow(-1), th, cell, tmp;  // table row, table header, table cell 	
  // create html table header row using the extracted headers above  
  for (let i in col) {
		if ( col[i] == '' ) { continue; } // set an empty string if you want to skip a column inside a 2-dim array
		th= document.createElement('th');      
		th.innerHTML = col[i];
		tr.appendChild(th);
  }
  if ( is2dArray(data) ) 
	{ // add 2-dim array data to the table as rows		
		tmp= data.length;
		for (let i = 0; i < tmp; i++) 
		{
			tr= table.insertRow(-1);
			for (let j in col) {
				if (col[j] == '') { continue; }  // set an empty string if you want to skip a column inside a 2-dim array
				cell= tr.insertCell(-1);
				cell.innerHTML = data[i][j];
			}
		}
  }
	else 
	{	// add 1-dim array data to the table as only one row
		tr= table.insertRow(-1);
		tmp=data.length;
		for (let i = 0; i < tmp; i++)	{
			cell= tr.insertCell(-1);
			cell.innerHTML = data[i];
		}
	}	
  elem.innerHTML= '';
  elem.appendChild(table); // finally add the newly created table with string data to a html container
}


// test if array is 2-dim (return true) or not (return false)
function is2dArray(arr)
{
	if(arr[0] === undefined)	{
		return false;
	}
	else	{
		return (arr[0].constructor === Array);
	}
}


// <!-- IMPORTANT: showOption() and hideOption() buttons must be outside form element -->
// save to window global vars all user Options/Settings
function hideOption()  
{
  setVarPar();   
  document.getElementById('options').style.display = 'none'; 
  document.getElementById('start').setAttribute('class', 'center-xy');   // ex 'child' class 
  document.getElementById('show_option').style.display = 'block';   
  document.getElementById('start_div').style.display = 'block';        //.removeAttribute("disabled");  
  document.getElementById('hide_option').style.display = 'none';   
  document.getElementById('instructions').style.display= 'block';
  refMsg(); 
  setIns();	
}


// begin of new code for export localStorage app data:  data= localStorage.getItem('myApp-v1.0');

// show form with application options
// <!-- IMPORTANT: showOption() and hideOption() buttons must be outside form element -->
function showOption()   
{
  document.getElementById('hide_option').style.display= 'block';
  document.getElementById('options').style.display= 'block';
  document.getElementById('start').classList.remove('center-xy');  // ex 'child' class    
  document.getElementById('start_div').style.display= 'none';    
  document.getElementById('show_option').style.display= 'none';  
  document.getElementById('instructions').style.display= 'none';     
  document.getElementById('store_oper').style.display= 'none';
	if (document.getElementById('store_group').disabled === false && localStorage.getItem('tde-v1.0')) 
  {
    let cnt= JSON.parse(localStorage.getItem('tde-v1.0')).length;                                                        
    document.getElementById('store_cnt').textContent=  'You have a total of ' + cnt  + ' TDE results stored in your web browser.';
    if (cnt > 0) {
			document.getElementById('store_oper').style.display= 'table';   // xxx
		}
	}
  dirty= false;
  document.getElementById('is_valid').textContent='';  // tmp reset validator message
}


// clear web browser storage data for this application
function clrStorage()  
{  
  let cnt= localStorage.getItem('tde-v1.0') ? JSON.parse(localStorage.getItem('tde-v1.0')).length : 0;                                                        
  if (cnt === 0)  {
    alert('You have no TDE data to delete');
    return;
  }  
  if ( !window.confirm('You have a total of ' + cnt  + ' TDE results stored in your web browser. Continue?') ) { 
   return;
 }                                             
 if (window.confirm('Delete total saved TDE results:  '  + cnt + '. This action cannot be undone. Final confirmation?') ) 
 {  
   window.localStorage.removeItem('tde-v1.0'); 
   document.getElementById('store_cnt').textContent= 'All data has been deleted. No data.';
	 document.getElementById('store_oper').style.display= 'none';
 }  
}


// copy data from user web browser to the Clipboard 
function copyStorage()  
{  
  let data= localStorage.getItem('tde-v1.0');
	if (!data)  {
    alert('You have no data to Copy');
    return;
  }  
	data = JSON.parse(data); 
	let arr= jsonToArray(data);	
	copy2DToClipboard(arr);
}	


// convert json object into 2D array
// data = JSON object
// return 2D array
function jsonToArray(data)  
{
   // extract column (fields) names  
  let total= data.length, col = [], row, res=[];  // res= result (2d array) 
  for (let i = 0; i < total; i++) {
    for (let key in data[i]) {
			if (col.indexOf(key) === -1)  {
				col.push(key);  // col= columns (fields) names
			}
    }
  } 
	res.push(col);  // add field names as 1 row
  let c = col.length;    // column count  
  // add JSON data to the 2d array as rows.
  for (let i = 0; i < total; i++) 
  {
		row= [];
    for (let j = 0; j < c; j++)  {     
      row.push(data[i][col[j]]);
    }
		res.push(row);
  }
	return res;	
}


// copy 2D array to Clipboard
function copy2DToClipboard(arr) {
  let csv = '', row, cell;
  for (row = 0; row < arr.length; row++) {
    for (cell = 0; cell < arr[row].length; cell++) {
      csv += (arr[row][cell]+'').replace(/[\n\t]+/g, ' ');
      if (cell+1 < arr[row].length) csv += '\t';
    }
    if (row+1 < arr.length) { 
			csv += '\n'; 
		}
  }
  copyTextToClipboard(csv);
}


// copy 2D array (or text) to Clipboard
function copyTextToClipboard(txt) 
{	
	if (!navigator.clipboard) {
    fallbackCopy(txt);
    return;
  }
  navigator.clipboard.writeText(txt).then(function() {
   //alert('Async: Copying to Clipboard was successful!');
  }, function(err) {
    alert('Oops, could not copy data. Error: ' + err);
  });
}


// fallback copy text to clipboard
function fallbackCopy(txt) 
{
  let ta = document.createElement('textarea');
  ta.value = txt; 
  // Avoid scrolling to bottom
  ta.style.top = '0';
  ta.style.left = '0';
  ta.style.position = 'fixed';
  document.body.appendChild(ta);
  ta.focus();
  ta.select();
  try {
    let successful = document.execCommand('copy');
    let msg = successful ? 'successful' : 'unsuccessful';
  } catch (err) {
    alert('Oops, unable to copy. Error: ' + err);
  }
  document.body.removeChild(ta);
}


// displays user saved results (Window.localStorage application data) in new empty web page (tab) 
function newWin()
{ 
	let data = JSON.parse(localStorage.getItem('tde-v1.0'))
	createTableJSON(document.getElementById('table_grp'), data);
	let stl= '<style>table{margin:1em auto;border:thick solid #000;border-collapse:collapse}td,th{border:medium solid #000;text-align:center;padding:.5em}th{font-weight:bold}</style>';
	let tbl= document.getElementById('table_grp').innerHTML;	
	let win = window.open('');
  win.document.write('<html><head><title>Saved cumulative data</title>' + stl + '</head><body>' + tbl + '</body></html>');
  win.focus();
}


// displays Latest research results in new empty web page (tab) 
function newWinLast()
{ 
  let ind = document.getElementById('last_data_sel').selectedIndex;
	ind++;
	let e = document.getElementById('table' + ind);	
	let stl= '<style>table{margin:1em auto;border:thick solid #000;border-collapse:collapse}td,th{border:medium solid #000;text-align:center;padding:.5em}th{font-weight:bold}</style>';
	let tbl= e.innerHTML; 	
	let win = window.open('');
	win.document.write('<html><head><title>Latest research data</title>' + stl + '</head><body>' + tbl + '</body></html>');
	win.focus();
}


// copy only latest task results (data) to the Clipboard 
function copyLast()  
{ 
  let ind = document.getElementById('last_data_sel').selectedIndex, tbl;
  switch (ind) {
    case 0:
			tbl = document.getElementById('table1').getElementsByTagName('table')[0];  // note: id='table1' is html div element
			break;		
    case 1:
			tbl = document.getElementById('table2').getElementsByTagName('table')[0];  // note: id='table2' is html div element
			break;
		default:
			tbl = document.getElementById('table3').getElementsByTagName('table')[0];  // note: id='table3' is html div element
			break;	
	}		
  if (tbl)  {
		let data= table2DArray(tbl);
		if (data.length > 0) {
			copy2DToClipboard(data);
		}
	}
}	


// input argument tbl == html/DOM  table  object
// return 2D array
function table2DArray(tbl) 
{
	let res = [];	 
	for (let i = 0, row; row = tbl.rows[i]; i++) 
	{   //iterate through rows. rows would be accessed using the "row" variable assigned in the for loop
	  res[i]= [];		
		for (let j = 0, col; col = row.cells[j]; j++) {  //iterate through columns. columns would be accessed using the "col" variable assigned in the for loop			 
		  res[i][j]= col.textContent;
		}  
	}
	return res;				
}


// https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Classes
// https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Classes/static
// http://www.rosettacode.org/mw/index.php?title=Special%3ASearch&search=integration&go=Go&fulltext=1
class Stat 
{ 
		
	// return arithmetic mean (mean, average) of input array (arr)
  // no validation => input array must be in valid number format	
	static m(arr)  
	{	
		let len= arr.length, sum= 0;
		for (let i=0; i<len; i++) {
			sum+= arr[i];
		}
		return (sum/arr.length);
	}	


	// return Sample Standard Deviation (with Bessel's correction) => N-1
	// no validation => input array must be in valid number format
	// arr = array argument;  mean = average for input arr 
	static sd(arr, mean) 
	{
		let len= arr.length, sum= 0, res;
		for (let i=0; i<len; i++) {
			sum+= Math.pow(arr[i]-mean, 2);	 // Math.pow(base, exponent)
		}
		res= Math.pow(sum/(len-1), 0.5);   // denominator = N-1
		return res;
	}
	
	
	// https://www.psychometrica.de/correlation.html
	// return Pearson correlation coefficient [ Pearson's r, Pearson product-moment correlation coefficient, bivariate correlation ]
	//  x, y = arrays 
	static r(x,y)
	{
		if (x.length != y.length) {
			return;  // undefined
		}
		let cov= 0, n, m1, m2, sd1, sd2;
		n= x.length-1;     
		m1= this.m(x);          //MeanAndStdDev(a, M1, SD1);
		sd1= this.sd(x,m1);
		m2= this.m(y);
		sd2= this.sd(y,m2);    //MeanAndStdDev(b, M2, SD2);		
		for (let i=0; i <= n; i++) {
			cov= ( x[i] - m1 ) * ( y[i] - m2 ) + cov;  // covariance
    }
		return ( cov / ( n * sd1 * sd2 ) );   // denominator = n-1 (Bessel's correction)		
	}
	
	
	// return covariance
  // x, y = arrays   
	static cov(x,y)
	{
		if (x.length != y.length) {
			return;  // undefined
		}
		let cov= 0, n, m1, m2, sd1, sd2;
		n= x.length-1;     
		m1= this.m(x);          
		sd1= this.sd(x,m1);
		m2= this.m(y);
		sd2= this.sd(y,m2);    		
		for (let i=0; i <= n; i++) {
			cov= ( x[i] - m1 ) * ( y[i] - m2 ) + cov  ;  // covarijanca
    }
		return cov;   // denominator = n-1 (Bessel's correction)		
	}


	// https://www.statisticshowto.com/probability-and-statistics/correlation-coefficient-formula/spearman-rank-correlation-definition-calculate/   
	// https://mathcracker.com/spearman-correlation-calculator
	// compute Spearman's rank correlation coefficient
	// x, y = arrays
	static r_spearman(x, y)
	{ 
		if (x.length != y.length) {
			return;  // undefined
		}
		const N = x.length;  // n is the number of observations
		let xr = x.slice();  // must be
		xr= this.rankify(xr, xr.length);   // get (tied) ranks for 1. array values
		
		let yr = y.slice();  // must be
		yr= this.rankify(yr, yr.length);	  // get (tied) ranks for 2. array values
		
		let mx= this.m(xr);  // mean rank for 1. array
		let my= this.m(yr);  // mean rank for 2. array
		
		let sxy= 0, sx= 0, sy= 0;
		for (let i= 0; i < N; i++)  
		{
			sxy+= ( xr[i] - mx ) * ( yr[i] - my); 
			sx+= Math.pow( ( xr[i] - mx ), 2);
			sy+= Math.pow( ( yr[i] - my ), 2);
		}
		sxy= sxy/N;
		sx= sx/N;
    sy= sy/N;
    let res = sxy / Math.sqrt(sx * sy);
		return res;	
	}
	
	
	// https://www.geeksforgeeks.org/rank-elements-array/
	// function to find tied ranks
	// A => input array;  n => input array length
	static rankify(A, n) 
	{	
		let R = [...Array(n)];  // Rank Vector
		// Sweep through all elements in A for each
		// element count the number of less than and
		// equal elements separately in r and s.
		for (let i = 0; i < n; i++) 
		{
			let r = 1, s = 1;
			for (let j = 0; j < n; j++) {
				if (j != i && A[j] < A[i]) r += 1;
				if (j != i && A[j] == A[i]) s += 1;
			}
			// Use formula to obtain rank
			R[i] = parseFloat(r + parseFloat(s - 1) / parseFloat(2));
		}
		//for (let i = 0; i < n; i++)  document.write(parseFloat(R[i]).toFixed(1) + "  ");
		return R;
	}
 
   
	// https://people.sc.fsu.edu/~jburkardt/cpp_src/asa005/asa005.cpp
	// https://people.sc.fsu.edu/~jburkardt/cpp_src/asa005/asa005.html
	// https://people.math.sc.edu/Burkardt/cpp_src/toms291/toms291.cpp
	// https://people.math.sc.edu/Burkardt/c_src/toms291/toms291.html


	//****************************************************************************80

	static alnorm ( x, upper )

	//****************************************************************************80
	//
	//  Purpose:
	//
	//    ALNORM computes the cumulative density of the standard normal distribution.
	//
	//  Licensing:
	//
	//    This code is distributed under the GNU LGPL license. 
	//
	//  Modified:
	//
	//    17 January 2008
	//
	//  Author:
	//
	//    Original FORTRAN77 version by David Hill.
	//    C++ version by John Burkardt.
	//    JS version by Kekec
	//  Reference:
	//
	//    David Hill,
	//    Algorithm AS 66:
	//    The Normal Integral,
	//    Applied Statistics,
	//    Volume 22, Number 3, 1973, pages 424-427.
	//
	//  Parameters:
	//
	//    Input, double X, is one endpoint of the semi-infinite interval
	//    over which the integration takes place.
	//
	//    Input, bool UPPER, determines whether the upper or lower
	//    interval is to be integrated:
	//    .TRUE.  => integrate from X to + Infinity;
	//    .FALSE. => integrate from - Infinity to X.
	//
	//    Output, double ALNORM, the integral of the standard normal
	//    distribution over the desired interval.
	//
	{
		let a1 = 5.75885480458;
		let a2 = 2.62433121679;
		let a3 = 5.92885724438;
		let b1 = -29.8213557807;
		let b2 = 48.6959930692;
		let c1 = -0.000000038052;
		let c2 = 0.000398064794;
		let c3 = -0.151679116635;
		let c4 = 4.8385912808;
		let c5 = 0.742380924027;
		let c6 = 3.99019417011;
		let con = 1.28;
		let d1 = 1.00000615302;
		let d2 = 1.98615381364;
		let d3 = 5.29330324926;
		let d4 = -15.1508972451;
		let d5 = 30.789933034;
		let ltone = 7.0;
		let p = 0.398942280444;
		let q = 0.39990348504;
		let r = 0.398942280385;
		let utzero = 18.66;
		let value;
		let y;
		let z = x;

		let up = upper;   // bool
		z = x;

		if ( z < 0.0 )
		{
			up = !up;
			z = - z;
		}

		if ( ltone < z && ( ( !up ) || utzero < z ) )
		{
			if ( up ) {
				value = 0;
			}
			else  {
				value = 1;
			}
			return value;
		}

		y = 0.5 * z * z;

		if ( z <= con )
		{
			value = 0.5 - z * ( p - q * y 
				/ ( y + a1 + b1 
				/ ( y + a2 + b2 
				/ ( y + a3 ))));
		}
		else
		{ 
			value = r * Math.exp( - y )   
				/ ( z + c1 + d1 
				/ ( z + c2 + d2 
				/ ( z + c3 + d3 
				/ ( z + c4 + d4 
				/ ( z + c5 + d5 
				/ ( z + c6 ))))));
		}

		if ( !up )  {
			value = 1 - value;
		}

		return value;
	}


	//****************************************************************************80

	static prncst ( st, idf, d, ifault )

	//****************************************************************************80
	//
	//  Purpose:
	//
	//    PRNCST computes the lower tail of noncentral T distribution.
	//
	//  Licensing:
	//
	//    This code is distributed under the GNU LGPL license. 
	//
	//  Modified:
	//
	//    26 January 2008
	//
	//  Author:
	//
	//    Original FORTRAN77 version by BE Cooper.
	//    C++ version by John Burkardt.
	//    JS version by Kekec
	//  Reference:
	//
	//    BE Cooper,
	//    Algorithm AS 5:
	//    The Integral of the Non-Central T-Distribution,
	//    Applied Statistics,
	//    Volume 17, Number 2, 1968, page 193.
	//
	//  Parameters:
	//
	//    Input, double ST, the argument.
	//
	//    Input, int IDF, the number of degrees of freedom.
	//
	//    Input, double D, the noncentrality parameter.
	//
	//    Output, int *IFAULT, error flag.
	//    0, no error occurred.
	//    nonzero, an error occurred.
	//
	//    Output, double PRNCST, the value of the lower tail of
	//    the noncentral T distribution.
	//
	//  Local Parameters:
	//
	//    Local, double G1, 1.0 / sqrt(2.0 * pi)
	//
	//    Local, double G2, 1.0 / (2.0 * pi)
	//
	//    Local, double G3, sqrt(2.0 * pi)
	//
	{
		let a, ak, b,  da, drb,  fk, fkm1, fmkm1, fmkm2, ioe, k, rb, sum, value;   
		let emin = 12.5;
		let g1 = 0.3989422804;
		let g2 = 0.1591549431;
		let g3 = 2.5066282746;
		let f = idf;   //( double ) ( idf );
	//
	//  For very large IDF, use the normal approximation.
	//
		if ( 100 < idf )
		{
			ifault = 1;

			a = Math.sqrt ( 0.5 * f ) * Math.exp ( this.lgamma ( 0.5 * ( f - 1 ) ) 
				- this.lgamma ( 0.5 * f ) ) * d;

			value = this.alnorm ( ( st - a ) / Math.sqrt ( f * ( 1 + d * d ) 
				/ ( f - 2 ) - a * a ), false );
			return value;
		}

		ifault = 0;
		ioe = ( idf % 2 );
		a = st / Math.sqrt ( f );
		b = f / ( f + st * st );
		rb = Math.sqrt ( b );
		da = d * a;
		drb = d * rb;

		if ( idf == 1 )
		{
			value = this.alnorm ( drb, true ) + 2 * this.tfn ( drb, a );
			return value;
		}

		sum = 0;

		if ( Math.abs ( drb ) < emin )
		{
			fmkm2 = a * rb * Math.exp ( - 0.5 * drb * drb ) 
			* this.alnorm ( a * drb, false ) * g1;
		}
		else
		{
			fmkm2 = 0;
		}

		fmkm1 = b * da * fmkm2;

		if ( Math.abs ( d ) < emin )
		{
			fmkm1 = fmkm1 + b * a * g2 * Math.exp ( - 0.5 * d * d );
		}

		if ( ioe == 0 )
		{
			sum = fmkm2;
		}
		else
		{
			sum = fmkm1;
		}

		ak = 1;
		fk = 2;

		for ( k = 2; k <= idf - 2; k = k + 2 )
		{
			fkm1 = fk - 1;
			fmkm2 = b * ( da * ak * fmkm1 + fmkm2 ) * fkm1 / fk;
			ak = 1 / ( ak * fkm1 );
			fmkm1 = b * ( da * ak * fmkm2 + fmkm1 ) * fk / ( fk + 1 );

			if ( ioe == 0 )
			{
				sum += fmkm2;
			}
			else
			{
				sum += fmkm1;
			}
			ak = 1 / ( ak * fk );
			fk += 2;
		}

		if ( ioe == 0 )
		{
			value = this.alnorm ( d, true ) + sum * g3;
		}
		else
		{
			value = this.alnorm ( drb, true ) + 2 * ( sum + this.tfn ( drb, a ) );
		}

		return value;
	}
	//****************************************************************************80


	// https://people.math.sc.edu/Burkardt/c_src/toms291/toms291.html
	// https://people.math.sc.edu/Burkardt/cpp_src/toms291/toms291.cpp
	//****************************************************************************80

	static lgamma ( x )

	//****************************************************************************80
	//
	//  Purpose:
	//
	//    ALOGAM computes the logarithm of the Gamma function.
	//
	//  Modified:
	//
	//    22 January 2008
	//
	//  Author:
	//
	//    Original FORTRAN77 version by Malcolm Pike, David Hill.
	//    C++ version by John Burkardt.
	//    JS version by Kekec
	//  Reference:
	//
	//    Malcolm Pike, David Hill,
	//    Algorithm 291:
	//    Logarithm of Gamma Function,
	//    Communications of the ACM,
	//    Volume 9, Number 9, September 1966, page 684.
	//
	//  Parameters:
	//
	//    Input, double X, the argument of the Gamma function.
	//    X should be greater than 0.
	//
	//    Output, double lgamma, the logarithm of the Gamma
	//    function of X.
	//
	{
		let f, value, y, z;

		if ( x <= 0 || isNaN(x) )  {
			return NaN;
		}

		y = x;

		if ( x < 7 )
		{
			f = 1;
			z = y;

			while ( z < 7 )
			{
				f = f * z;
				z++; 
			}
			y = z;
			f = - Math.log ( f );
		}
		else
		{
			f = 0;
		}

		z = 1 / y / y;

		value = f + ( y - 0.5 ) * Math.log ( y ) - y 
			+ 0.918938533204673 + 
			((( 
			- 0.000595238095238   * z 
			+ 0.000793650793651 ) * z 
			- 0.002777777777778 ) * z 
			+ 0.083333333333333 ) / y;

		return value;
	}

	//****************************************************************************80

	static tfn ( x, fx )

	//****************************************************************************80
	//
	//  Purpose:
	//
	//    TFN calculates the T-function of Owen.
	//
	//  Licensing:
	//
	//    This code is distributed under the GNU LGPL license. 
	//
	//  Modified:
	//
	//    16 January 2008
	//
	//  Author:
	//
	//    Original FORTRAN77 version by JC Young, Christoph Minder.
	//    C++ version by John Burkardt.
	//    JS version by Kekec
	//  Reference:
	//
	//    MA Porter, DJ Winstanley,
	//    Remark AS R30:
	//    A Remark on Algorithm AS76:
	//    An Integral Useful in Calculating Noncentral T and Bivariate
	//    Normal Probabilities,
	//    Applied Statistics,
	//    Volume 28, Number 1, 1979, page 113.
	//
	//    JC Young, Christoph Minder,
	//    Algorithm AS 76: 
	//    An Algorithm Useful in Calculating Non-Central T and 
	//    Bivariate Normal Distributions,
	//    Applied Statistics,
	//    Volume 23, Number 3, 1974, pages 455-457.
	//
	//  Parameters:
	//
	//    Input, double X, FX, the parameters of the function.
	//
	//    Output, double TFN, the value of the T-function.
	//
	{
		const NG = 5;

		let fxs, i, r1, r2, rt, value, x1, x2, xs;
		let r = [
			0.1477621, 
			0.1346334, 
			0.1095432, 
			0.0747257, 
			0.0333357 ];
		let tp = 0.159155;
		let tv1 = 1.0E-35;
		let tv2 = 15.0;
		let tv3 = 15.0;
		let tv4 = 1.0E-05;
		let u = [
			0.0744372, 
			0.2166977, 
			0.3397048,
			0.4325317, 
			0.4869533 ];


	//
	//  Test for X near zero.
	//
		if ( Math.abs ( x ) < tv1 )
		{
			value = tp * Math.atan ( fx );
			return value;
		}
	//
	//  Test for large values of abs(X).
	//
		if ( tv2 < Math.abs ( x ) )
		{
			value = 0.0;
			return value;
		}
	//
	//  Test for FX near zero.
	//
		if ( Math.abs ( fx ) < tv1 )
		{
			value = 0;
			return value;
		}
	//
	//  Test whether abs ( FX ) is so large that it must be truncated.
	//
		xs = - 0.5 * x * x;
		x2 = fx;
		fxs = fx * fx;
	//
	//  Computation of truncation point by Newton iteration.
	//
		if ( tv3 <= Math.log ( 1 + fxs ) - xs * fxs )
		{
			x1 = 0.5 * fx;
			fxs = 0.25 * fxs;

			for ( ; ; )
			{
				rt = fxs + 1;

				x2 = x1 + ( xs * fxs + tv3 - Math.log ( rt ) ) 
				/ ( 2 * x1 * ( 1 / rt - xs ) );

				fxs = x2 * x2;

				if ( Math.abs ( x2 - x1 ) < tv4 )  {
					break;
				}
				x1 = x2;
			}
		}
	//
	//  Gaussian quadrature.
	//
		rt = 0.0;
		for ( i = 0; i < NG; i++ )
		{
			r1 = 1 + fxs * Math.pow ( 0.5 + u[i], 2 );
			r2 = 1 + fxs * Math.pow ( 0.5 - u[i], 2 );

			rt = rt + r[i] * ( Math.exp ( xs * r1 ) / r1 + Math.exp ( xs * r2 ) / r2 );
		}

		value = rt * x2 * tp;

		return value;

	}
	//****************************************************************************80
	
} 